// Event.h: interface for the TEvent class (speed optimized Event class).
//

#pragma once

#undef DECL
#if defined(_AFXDLL) || defined(XPERT_EXE)
#ifdef UTILS_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif
#else
#define DECL 
#endif

//
// TEvent is a class designed to improve the performance of events when used as flags in a processing loop.
// In Windows CE 6, but even in Windows CE 5 there is a large performance penalty to enter the kernel and check
// the status of an event flag by calling WaitForSingleObject. This sort of code can be sped up greatly by
// using a boolean flag. TEvent combines the two concepts by providing an event that can be checked almost as
// fast as a boolean. I'm calling this a "smart" event. TEvent can also encapsulate a "dumb" events in order
// to facilitates mixing smart and dumb events when calling WaitForMultiple objects (see TEventArray).
//
// WARNING: A crucial aspect of TEvent is that the underlying event HANDLE is never exposed. If it was, then
//          it would be very easy for user code to get the underlying boolean flag out of sync with the event.
//
// A smart event has extra code that tracks the current state of the underlying event HANDLE  such that whenever 
// posible this flag is checked rather than having to make an expensive call in to the Kernel to retrieve the state via
// ::WaitForSingleObject().
//
// A dumb event is an event HANDLE that must be used as-is and cannot be optimized because it's
// a named events, shared across threads, or might be manipulated by other code by it's HANDLE.
// Why bother? Since it's impossible to access the handle of a TEvent the only way to wait on a mix of 
// both smart and dumb events is to encapsulate the dumb events in a TEventHandle. In fact this is done
// automatically when building up an array of events to pass to ::WaitForMultipleObjects():
//
// Given a smart event and a dumb event handle:
//
//    TEvent SmartEvent(TRUE, FALSE);
//    HANDLE hDumbEvent = CreateEvent(0, TRUE, FALSE, L"Named Event Allowed");
//
// We might want to do something like this, but we cannot because the Smart event could become out of sync:
//
//    HANDLE handles[] = {SmartEvent.GetHandleNotAllowed(), hDumbEvent};
//    DWORD result = ::WaitForMultipleObjects(0, handles, FALSE, 60000)
//
// Here's what we should do, instead:
//
//    TEventArray handles[] = {SmartEvent, hDumbEvent};
//    DWORD result = TEvent::WaitForMultipleObjects(0, handles, FALSE, 60000)
//
// This works because C++ will automatically convert a HANDLE in to a TEventArray as needed.
//
// If you're wondering why you can't just create an array of TEvent's rather than use TEventArray, it's because
// C++ does not support arrays of objects!!! Which is not to say it will prevent you from doing so, but it will
// not work as you might expect. Arrays need to be used with simple scalar data types that can be raw copied. TEventArray 
// works because it stores a pointer to an event (and a boolean flag), rather than the actual event object. 
//


class DECL TEventArray;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TEvent encapsulates simple events that avoid slow system calls where possible by keeping track of the state of the event
// This can payoff in a tight processing loop where the calls to check an event to determine the end of the loop could be far
// more expensive than the processing inside the loop (ex: Basic interpreter).
//
// Notes:
// - This means the all code which manipulates this event does so through the TEvent class (the event handle is not exposed and
//   should never be exposed to avoid any situations that could violate this import pre-condition)
// - IsSet() and IsReset() will not clear the state of a manual reset event, use Wait(0) instead
// - Named events are not supported, but they can be encapsulated with TEventHandle if desired
//
class DECL TEvent
{
public:
   TEvent(BOOL bManualReset, BOOL bInitialState);
   TEvent(HANDLE hEvent);
   ~TEvent();
   bool IsSet() const;
   bool IsReset() const;
   BOOL AutoReset();
   BOOL Set();
   BOOL Reset();
   DWORD Wait(DWORD dwTimeout);
   static DWORD WaitForMultiple(DWORD cEvents, const TEventArray* pEvents, BOOL fWaitAll, DWORD dwTimeout);
   BOOL IsSmart() const                   { return m_bSmart;                                             };
private:
   friend class TEventArray;
   void AddRef();
   bool DelRef();
   HANDLE GetHandle() const               { return m_hEvent;                                             };
   HANDLE m_hEvent;
   BOOL m_bSmart;
   volatile BOOL m_bEvent;
   BOOL m_bAutoReset;
   CRITICAL_SECTION m_csLock;
   LONG m_iReferenceCount;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TEventArray is used to create an array of TEvent's that can be used in calls to TEvent::WaitForMultiple()
// For best performance, dumb events should come after smart events in a TEventArray[], albeit priority is
// determined by order - so this might not always be possible.
//
// Technically speaking, when we write something like this:
//
//    TEventArray events[] = {eventA, eventB, hEventC};
//
// We're creating an array of TEventArray objects, rather than array of TEvent's but TEventArray is designed to make this as
// transparent as possible by providing accessors that make it function like a TEvent. 
//
class DECL TEventArray
{
public:
   TEventArray()                       { m_pEvent = NULL; m_bDelete = false;                                                  };
   TEventArray(const TEvent& Event)    { m_pEvent = (TEvent *) &Event; m_bDelete = false;                                     };
   TEventArray(const TEvent* Event)    { m_pEvent = (TEvent *) Event; m_bDelete = false;                                      };
   TEventArray(BOOL bManualReset, BOOL bInitialState);
   TEventArray(HANDLE hEvent);
   ~TEventArray();
   TEventArray(const TEventArray& Array);
   const TEventArray& operator=(const TEventArray& Array);
   bool IsSet() const                  { return m_pEvent->IsSet();                                                            };
   bool IsReset() const                { return m_pEvent->IsReset();                                                          };
   BOOL Set() const                    { return m_pEvent->Set();                                                              };
   BOOL Reset() const                  { return m_pEvent->Reset();                                                            };
   DWORD Wait(DWORD dwTimeout) const   { return m_pEvent->Wait(dwTimeout);                                                    };
   BOOL IsSmart() const                { return m_pEvent->IsSmart();                                                          };
   BOOL AutoReset() const              { return m_pEvent->AutoReset();                                                        };
protected:
   friend class TEvent;
   friend class TVariantEvent;
   HANDLE GetHandle() const            { return m_pEvent->GetHandle();                                                        };
   TEvent* m_pEvent;
   bool m_bDelete;
};

inline TEventArray::TEventArray(BOOL bManualReset, BOOL bInitialState) 
{ 
   m_pEvent = new TEvent(bManualReset, bInitialState);
   m_bDelete = true;
   m_pEvent->AddRef();
}

inline TEventArray::TEventArray(HANDLE hEvent)
{
   m_pEvent = new TEvent(hEvent);
   m_bDelete = true;
   m_pEvent->AddRef();
}

inline TEventArray::~TEventArray()
{
   if (m_bDelete)
      if (m_pEvent->DelRef())
         delete m_pEvent;
};

inline TEventArray::TEventArray(const TEventArray& Array)
{
   m_pEvent = Array.m_pEvent;
   m_bDelete = Array.m_bDelete;
   if (m_bDelete)
      m_pEvent->AddRef();
}

inline const TEventArray& TEventArray::operator=(const TEventArray& Array)
{
   m_pEvent = Array.m_pEvent;
   m_bDelete = Array.m_bDelete;
   if (m_bDelete)
      m_pEvent->AddRef();
   return *this;
}
